package lcsb.mapviewer.api.projects;

import java.awt.geom.Path2D;
import java.awt.geom.PathIterator;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.io.IOUtils;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.bind.annotation.CookieValue;
import org.springframework.web.bind.annotation.RequestParam;

import lcsb.mapviewer.api.BaseRestImpl;
import lcsb.mapviewer.api.QueryException;
import lcsb.mapviewer.api.projects.models.publications.PublicationsRestImpl;
import lcsb.mapviewer.commands.ClearColorModelCommand;
import lcsb.mapviewer.commands.ColorExtractor;
import lcsb.mapviewer.commands.ColorModelCommand;
import lcsb.mapviewer.commands.CommandExecutionException;
import lcsb.mapviewer.commands.CopyCommand;
import lcsb.mapviewer.commands.SubModelCommand;
import lcsb.mapviewer.common.Configuration;
import lcsb.mapviewer.converter.ConverterException;
import lcsb.mapviewer.converter.IConverter;
import lcsb.mapviewer.converter.graphics.AbstractImageGenerator.Params;
import lcsb.mapviewer.converter.graphics.ImageGenerators;
import lcsb.mapviewer.converter.model.celldesigner.CellDesignerXmlParser;
import lcsb.mapviewer.converter.model.sbgnml.SbgnmlXmlConverter;
import lcsb.mapviewer.model.Project;
import lcsb.mapviewer.model.cache.FileEntry;
import lcsb.mapviewer.model.cache.UploadedFileEntry;
import lcsb.mapviewer.model.map.InconsistentModelException;
import lcsb.mapviewer.model.map.OverviewImage;
import lcsb.mapviewer.model.map.OverviewImageLink;
import lcsb.mapviewer.model.map.OverviewLink;
import lcsb.mapviewer.model.map.layout.ColorSchema;
import lcsb.mapviewer.model.map.layout.InvalidColorSchemaException;
import lcsb.mapviewer.model.map.layout.Layout;
import lcsb.mapviewer.model.map.model.Model;
import lcsb.mapviewer.model.map.species.Element;
import lcsb.mapviewer.model.user.User;
import lcsb.mapviewer.services.SecurityException;
import lcsb.mapviewer.services.interfaces.ILayoutService;
import lcsb.mapviewer.services.interfaces.IModelService;
import lcsb.mapviewer.services.interfaces.IProjectService;
import lcsb.mapviewer.services.interfaces.IUserService;
import lcsb.mapviewer.services.utils.ColorSchemaReader;
import lcsb.mapviewer.services.utils.data.BuildInLayout;
import lcsb.mapviewer.services.utils.gmap.CoordinationConverter;
import lcsb.mapviewer.services.view.AnnotationViewFactory;
import lcsb.mapviewer.services.view.AuthenticationToken;
import lcsb.mapviewer.services.view.OverviewImageViewFactory;

@Transactional(value = "txManager")
public class ProjectRestImpl extends BaseRestImpl {

	/**
	 * Constant defining size of the array returned by
	 * {@link PathIterator#currentSegment(double[])} method. More nformation can
	 * be found <a href=
	 * "http://docs.oracle.com/javase/7/docs/api/java/awt/geom/PathIterator.html#currentSegment(double[])"
	 * >here</a>
	 */
	private static final int				 PATH_ITERATOR_SEGMENT_SIZE	= 6;

	Logger													 logger											= Logger.getLogger(ProjectRestImpl.class);

	@Autowired
	private IUserService						 userService;

	@Autowired
	private IProjectService					 projectService;

	@Autowired
	private IModelService						 modelService;

	@Autowired
	private PublicationsRestImpl		 publicationsRestImpl;

	@Autowired
	private AnnotationViewFactory		 annotationViewFactory;

	@Autowired
	private ILayoutService					 layoutService;

	@Autowired
	private OverviewImageViewFactory factory;

	public ProjectMetaData getMetaData(String projectId, String token) throws SecurityException {
		AuthenticationToken authenticationToken = userService.getToken(token);
		Project project = projectService.getProjectByProjectId(projectId, authenticationToken);
		ProjectMetaData result = createData(project, authenticationToken);
		if (project.getOrganism() != null) {
			result.setOrganism(annotationViewFactory.create(project.getOrganism()));
		}
		if (project.getDisease() != null) {
			result.setDisease(annotationViewFactory.create(project.getDisease()));
		}
		return result;
	}

	private ProjectMetaData createData(Project project, AuthenticationToken token) {
		ProjectMetaData result = new ProjectMetaData();
		Model model = modelService.getLastModelByProjectId(project.getProjectId(), token);

		result.setName(project.getName());
		result.setProjectId(project.getProjectId());
		result.setIdObject(project.getId());

		if (model != null) {
			result.setOverviewImageViews(factory.createList(model.getOverviewImages()));
			result.setVersion(model.getMapVersion());
			result.setDescription(model.getNotes());

			Set<OverviewImage> set = new HashSet<OverviewImage>();
			set.addAll(model.getOverviewImages());
			for (OverviewImage image : model.getOverviewImages()) {
				for (OverviewLink ol : image.getLinks()) {
					if (ol instanceof OverviewImageLink) {
						set.remove(((OverviewImageLink) ol).getLinkedOverviewImage());
					}
				}
			}
			if (set.size() > 0) {
				result.setTopOverviewImage(factory.create(set.iterator().next()));
			} else if (model.getOverviewImages().size() > 0) {
				logger.warn("Cannot determine top level image. Taking first one. " + model.getOverviewImages().get(0).getFilename());
				result.setTopOverviewImage(factory.create(model.getOverviewImages().get(0)));
			}
			result.setMap(new ModelMetaData(model));

			List<Model> models = new ArrayList<>();
			models.add(model);
			models.addAll(model.getSubmodels());
			result.setPublicationCount(publicationsRestImpl.getPublications(models).size());

		}

		return result;
	}

	/**
	 * @return the userService
	 * @see #userService
	 */
	public IUserService getUserService() {
		return userService;
	}

	/**
	 * @param userService
	 *          the userService to set
	 * @see #userService
	 */
	public void setUserService(IUserService userService) {
		this.userService = userService;
	}

	/**
	 * @return the modelService
	 * @see #modelService
	 */
	public IModelService getModelService() {
		return modelService;
	}

	/**
	 * @param modelService
	 *          the modelService to set
	 * @see #modelService
	 */
	public void setModelService(IModelService modelService) {
		this.modelService = modelService;
	}

	/**
	 * @return the factory
	 * @see #factory
	 */
	public OverviewImageViewFactory getFactory() {
		return factory;
	}

	/**
	 * @param factory
	 *          the factory to set
	 * @see #factory
	 */
	public void setFactory(OverviewImageViewFactory factory) {
		this.factory = factory;
	}

	public FileEntry getSource(String token, String projectId) throws SecurityException, QueryException {
		AuthenticationToken authenticationToken = userService.getToken(token);
		Project project = projectService.getProjectByProjectId(projectId, authenticationToken);
		if (project == null) {
			throw new QueryException("Project with given id doesn't exist");
		}
		FileEntry entry = new UploadedFileEntry();
		entry.setOriginalFileName(project.getInputFileName());
		entry.setFileContent(project.getInputData());
		return entry;
	}

	public FileEntry getModelAsImage(String token, String projectId, String modelId, String handlerClass, String backgroundOverlayId, String overlayIds,
			String zoomLevel, String polygonString) throws SecurityException, QueryException, IOException, InvalidColorSchemaException, CommandExecutionException {
		AuthenticationToken authenticationToken = userService.getToken(token);
		User user = userService.getUserByToken(authenticationToken);

		Model topModel = modelService.getLastModelByProjectId(projectId, authenticationToken);
		if (topModel == null) {
			throw new QueryException("Project with given id doesn't exist");
		}

		Model originalModel = topModel.getSubmodelById(modelId);

		if (originalModel == null) {
			throw new QueryException("Model with given id doesn't exist");
		}

		Layout layout = null;
		if (!backgroundOverlayId.equals("")) {
			layout = topModel.getLayoutByIdentifier(Integer.valueOf(backgroundOverlayId));

			if (layout == null) {
				throw new QueryException("Unknown layout in model. Layout.id=" + backgroundOverlayId);
			}
		}

		Model colorModel = new CopyCommand(originalModel).execute();
		if (layout != null) {
			if (layout.getInputData() != null) {
				ColorSchemaReader reader = new ColorSchemaReader();
				Collection<ColorSchema> schemas = reader.readColorSchema(layout.getInputData().getFileContent());

				new ColorModelCommand(colorModel, schemas, userService.getColorExtractorForUser(user)).execute();
			} else if (layout.getTitle().equals(BuildInLayout.CLEAN.getTitle())) {
				// this might not return true if we change CLEAN.title in future...

				// if it's clean then remove coloring
				new ClearColorModelCommand(colorModel).execute();
			}
		}
		for (Element alias : colorModel.getElements()) {
			alias.setVisibilityLevel(0);
		}

		Integer level = Configuration.MIN_ZOOM_LEVEL;
		if (!zoomLevel.equals("")) {
			level = Integer.valueOf(zoomLevel);
		}

		// transform polygon
		CoordinationConverter cc = new CoordinationConverter(colorModel);
		Double minX = originalModel.getWidth();
		Double minY = originalModel.getHeight();
		Double maxX = 0.0;
		Double maxY = 0.0;
		if (!polygonString.equals("")) {
			Path2D polygon = cc.latLngToPolygon(polygonString);

			PathIterator pathIter = polygon.getPathIterator(null);
			while (!pathIter.isDone()) {
				final double[] segment = new double[PATH_ITERATOR_SEGMENT_SIZE];
				if (pathIter.currentSegment(segment) != PathIterator.SEG_CLOSE) {
					minX = Math.min(minX, segment[0]);
					maxX = Math.max(maxX, segment[0]);
					minY = Math.min(minY, segment[1]);
					maxY = Math.max(maxY, segment[1]);
				}
				pathIter.next();
			}
		} else {
			maxX = originalModel.getWidth();
			maxY = originalModel.getHeight();
			minX = 0.0;
			minY = 0.0;
		}

		maxX = Math.min(originalModel.getWidth(), maxX);
		maxY = Math.min(originalModel.getHeight(), maxY);
		minX = Math.max(0.0, minX);
		minY = Math.max(0.0, minY);

		Double scale = Math.max(originalModel.getHeight(), originalModel.getWidth()) / (originalModel.getTileSize());

		for (int i = level; i > Configuration.MIN_ZOOM_LEVEL; i--) {
			scale /= 2;
		}

		ColorExtractor colorExtractor = userService.getColorExtractorForUser(user);

		Params params = new Params().//
				x(minX).//
				y(minY).//
				height((maxY - minY) / scale).//
				width((maxX - minX) / scale).//
				level(level).//
				nested(false).// automatically set nested view as invalid
				scale(scale).//
				minColor(colorExtractor.getMinColor()).//
				maxColor(colorExtractor.getMaxColor()).//
				sbgn(topModel.getProject().isSbgnFormat()).//
				model(colorModel);
		List<Integer> visibleLayoutIds = deserializeIdList(overlayIds);
		for (Integer integer : visibleLayoutIds) {
			Map<Object, ColorSchema> map = layoutService.getElementsForLayout(colorModel, integer, authenticationToken);
			params.addVisibleLayout(map);
		}

		ImageGenerators imageGenerator = new ImageGenerators();
		String extension = imageGenerator.getExtension(handlerClass);
		File file = File.createTempFile("map", "." + extension);

		imageGenerator.generate(handlerClass, params, file.getAbsolutePath());

		FileEntry entry = new UploadedFileEntry();
		entry.setOriginalFileName("map." + extension);
		entry.setFileContent(IOUtils.toByteArray(new FileInputStream(file)));
		file.delete();
		return entry;

	}

	private List<Integer> deserializeIdList(String overlayIds) {
		List<Integer> result = new ArrayList<>();
		String[] tmp = overlayIds.split(",");
		for (String string : tmp) {
			if (!string.equals("")) {
				result.add(Integer.valueOf(string));
			}
		}
		return result;
	}

	public FileEntry getModelAsModelFile(String token, String projectId, String modelId, String handlerClass, String backgroundOverlayId, String overlayIds,
			String zoomLevel, String polygonString) throws SecurityException, QueryException, IOException, InvalidColorSchemaException, CommandExecutionException,
			ConverterException, InconsistentModelException {
		AuthenticationToken authenticationToken = userService.getToken(token);

		Model topModel = modelService.getLastModelByProjectId(projectId, authenticationToken);
		if (topModel == null) {
			throw new QueryException("Project with given id doesn't exist");
		}

		Model originalModel = topModel.getSubmodelById(modelId);

		if (originalModel == null) {
			throw new QueryException("Model with given id doesn't exist");
		}

		CoordinationConverter cc = new CoordinationConverter(originalModel);
		Path2D polygon = cc.latLngToPolygon(polygonString);

		// create model bounded by the polygon
		SubModelCommand subModelCommand = new SubModelCommand(originalModel, polygon);
		Model part = subModelCommand.execute();

		IConverter parser;
		if (SbgnmlXmlConverter.class.getCanonicalName().equals(handlerClass)) {
			parser = new SbgnmlXmlConverter();
		} else if (CellDesignerXmlParser.class.getCanonicalName().equals(handlerClass)) {
			parser = new CellDesignerXmlParser();
		} else {
			throw new QueryException("Unknown handlerClass: " + handlerClass);
		}
		InputStream is = parser.exportModelToInputStream(part);

		String fileExtension = parser.getFileExtension();

		FileEntry entry = new UploadedFileEntry();
		entry.setOriginalFileName("model." + fileExtension);
		entry.setFileContent(IOUtils.toByteArray(is));
		return entry;

	}

}
